/*
https://github.com/tttapa/ESP8266
+ weigu.lu
*/

#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>
#include <WebSocketsServer.h>
#include <Adafruit_NeoPixel.h>
#include "html_css_js.h"

#define DEBUG

ESP8266WebServer http_server(80); // create a web server on port 80
WebSocketsServer ws_server(81);   // create a ws server on port 81

const byte NEOPIX_PIN     = D2;
const byte NUM_PIX_RING_1 = 12;
const byte NUM_PIX_RING_2 = 24;
const byte NUM_PIX_RINGS  = 36;

const byte RING  = 3;             //1 = inner (12) 2 = outer(24), 3 = both

Adafruit_NeoPixel pixels = Adafruit_NeoPixel(NUM_PIX_RINGS, NEOPIX_PIN,
                           NEO_GRBW + NEO_KHZ800);

const char *ssid = "neo_lamp";         // SSID
const char *password = "mysecretpassword";      // password
int channel = 5;
IPAddress IP_AP(192, 168, 168, 168);     // fixed IP 
IPAddress MASK_AP(255, 255, 255, 0);

bool rainbow = false;                    // rainbow effect off on startup

unsigned long prevMillis = millis();
int hue = 0;

/******** SETUP ***************************************************************/
void setup() {
  pixels.setBrightness(255); // full brightness (255) only with LED power supply
  pixels.begin();            // initialise the NeoPixel library.
  pixels.show();             // Pixel off
  
#ifdef DEBUG
  Serial.begin(115200);      // Debug
  delay(10);
  Serial.println("\r\n");  
  Serial.setDebugOutput(true);
#endif // DEBUG
  setallpixels(RING,120,120,120,120);
  init_wifi_ap();            // initialise wifi ap
  init_ws_server();          // init websocket_server server  
  init_http_server();        // init a http server 
}

/***** MAINLOOP ***************************************************************/

void loop() {
  ws_server.loop();             // constantly check for websocket_server events
  http_server.handleClient();   // run the http server  
  if(rainbow) {                 // if the rainbow effect is turned on
    if(millis() > prevMillis + 20) {          
      if(++hue == 360)          // cycle through color wheel (1 degree/20ms)
        hue = 0;
      setHue(hue);              // set the RGB LED to the right color
      prevMillis = millis();
    }
  }
}
/***** INIT functions *********************************************************/

// initialise wifi ap
void init_wifi_ap() { 
  WiFi.softAPdisconnect(true);
  WiFi.disconnect(true);
  delay(200);
  WiFi.mode(WIFI_AP);  
  WiFi.softAP(ssid, password,channel);              // start the access point
  WiFi.softAPConfig(IP_AP, IP_AP, MASK_AP); // second is gateway same as ip
  IPAddress myIP = WiFi.softAPIP();
  #ifdef DEBUG  
    Serial.print("Access Point \"");
    Serial.print(ssid);
    Serial.println("\" started\r\n");
    Serial.print("IP address: ");
    Serial.println(myIP);
  #endif //DEBUG  
}

// initialise websocket_server server
void init_ws_server() {                       
  ws_server.begin();                  // start the ws server
  ws_server.onEvent(ws_server_event); // if incomming ws message,
                                      // do websocket_server_event()
  #ifdef DEBUG  
    Serial.println("websocket_server server started");
  #endif //DEBUG  
}

// initialise a http server 
void init_http_server() {   
  // handle http requests; root must be called! type: http://192.168.168.168/
  http_server.onNotFound(handle_not_found);     // 
  http_server.on("/", handle_root);              // (don't forget the "/")
  
  http_server.begin();                             // start the HTTP server
  http_server.setContentLength(html_css_js.length()); // if not given we get
                                            // net:err_content_length_mismatch
  #ifdef DEBUG  
    Serial.println("http server started.");
  #endif //DEBUG   
}
/***** SERVER handlers ********************************************************/

// if the requested page doesn't exist, return a 404 not found error
void handle_not_found() {   
  http_server.send(404, "text/plain", "404: Not Found");
}

// handle root 
void handle_root() {   
  http_server.send(200,"text/html",html_css_js); // HTTP response code 200
}
// do this when a websocket_server message is received
void ws_server_event(byte num, WStype_t type, byte * payload, size_t lenght) { 
  switch (type) {
    case WStype_DISCONNECTED:         // if the websocket_server is disconnected
      #ifdef DEBUG  
        Serial.printf("[%u] Disconnected!\n", num);
      #endif //DEBUG   
      break;
     // if a new websocket_server connection is established  
    case WStype_CONNECTED: {             
        IPAddress ip = ws_server.remoteIP(num);
        #ifdef DEBUG  
          Serial.printf("[%u] Connected from %d.%d.%d.%d url: %s\n", num, ip[0], ip[1], ip[2], ip[3], payload);
        #endif //DEBUG     
        rainbow = false; // rainbow off when a new connection is established
      }
      break;
    case WStype_TEXT:                     // if new text data is received
      Serial.printf("[%u] get Text: %s\n", num, payload);
      if (payload[0] == '#') {            // we get RGB data
        // decode rgbw data
        uint32_t rgbw = (uint32_t) strtol((const char *) &payload[1], NULL, 16);  
        int r = ((rgbw >> 24) & 0xFF);    // 8 bits per color, so R: bits 24-31
        int g = ((rgbw >> 16) & 0xFF);    // G: bits 16-23
        int b = ((rgbw >> 8)  & 0xFF);    // G: bits 8-15
        int w = (        rgbw & 0xFF);    // W: bits 0-7        
        #ifdef DEBUG  
          Serial.printf("Red: %d\n", r);
          Serial.printf("Green: %d\n", g);
          Serial.printf("Blue: %d\n", b);
          Serial.printf("White: %d\n", w);
        #endif //DEBUG   
        setallpixels(RING,r,g,b,w);
      } else if (payload[0] == 'R') { // browser sends R when rainbow effect
        rainbow = true;
      } else if (payload[0] == 'N') { // browser sends N when no rainbow effect
        rainbow = false;
      }
      break;
  }
}
/***** HELPER functions********************************************************/
// Set the RGB LED to a given hue (color) (0° = Red, 120° = Green, 240° = Blue)
void setHue(int hue) { 
  hue %= 360;                   // hue is an angle between 0 and 359°
  float radH = hue*3.142/180;   // Convert degrees to radians
  float rf, gf, bf;
  if(hue>=0 && hue<120){        // Convert from HSI color space to RGB              
    rf = cos(radH*3/4);
    gf = sin(radH*3/4);
    bf = 0;
  } else if(hue>=120 && hue<240){
    radH -= 2.09439;
    gf = cos(radH*3/4);
    bf = sin(radH*3/4);
    rf = 0;
  } else if(hue>=240 && hue<360){
    radH -= 4.188787;
    bf = cos(radH*3/4);
    rf = sin(radH*3/4);
    gf = 0;
  }
  int r = rf*rf*255;
  int g = gf*gf*255;
  int b = bf*bf*255;
  setallpixels(RING,r,g,b,0);
}  


void setallpixels(byte RING, byte R, byte G, byte B, byte W) {
  if (RING == 1) {
    for(int i=0;i<NUM_PIX_RING_1;i++) {
      pixels.setPixelColor(i, pixels.Color(R,G,B,W));
    }
  }
  else if (RING == 2) {
    for(int i=NUM_PIX_RING_2;i<NUM_PIX_RINGS;i++) {
      pixels.setPixelColor(i, pixels.Color(R,G,B,W));
    }
  }
  else if (RING == 3) {
    for(int i=0;i<NUM_PIX_RINGS;i++) {
      pixels.setPixelColor(i, pixels.Color(R,G,B,W));
    }
  }
  pixels.show();
}
